/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/






#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

//created on 09/15/98 by Igor Markov from an fmMoveManager.h
//by Andrew Caldwell and Igor Markov (06/02/98)
#if (!defined(_MSC_VER) || defined(_INSIDE_FMMOVEMANAGER_H) )

#ifndef _FMMMbase_CXX_
#define _FMMMbase_CXX_

#ifndef _INSIDE_FMMOVEMANAGER_H
#include "fmMoveManager.h"
#endif // !defined(_INSIDE_FMMOVEMANAGER_H)

#include <algorithm>
using std::fill;
#include "PartEvals/netCut2way.h"

template<class Evaluator, class GainContainer>
FMMMbase<Evaluator,GainContainer> ::FMMMbase(const PartitioningProblem& problem,
					     bool allowCorkingNodes)
        :MoveManagerTemplateBase<Evaluator, GainContainer>(problem),
         _lockedModules(problem.getHGraph().getNumNodes()),
         _lockedNets        (problem.getHGraph().getNumEdges()),
         _lockedNetConfigIds(problem.getHGraph().getNumEdges()),
         _numMovable(problem.getHGraph().getNumNodes() -
                                      problem.getHGraph().getNumTerminals()),
         _numTerminals(problem.getHGraph().getNumTerminals()),
         _numParts(problem.getNumPartitions()),
         _moveLog(), _movingModuleIdx(UINT_MAX), _moveFromPart(UINT_MAX), 
	 _moveToPart(UINT_MAX), 
         _numIllegalMoves(0), _allowCorkingNodes(allowCorkingNodes)
{  
   // need to call resetTo before the pass and then
   // computeGains()  - partitoner needs to call this when it is ready;
   //                   this will unlock the nodes that are being considered
   const vector<vector<double> >& maxAreas = problem.getMaxCapacities();
   const vector<vector<double> >& minAreas = problem.getMinCapacities();
   _tol = min(maxAreas[0][0]-minAreas[0][0], maxAreas[1][0]-minAreas[1][0]);

}

template<class Evaluator, class GainContainer>
void FMMMbase<Evaluator,GainContainer>::initializeGainBuckets()
{
    unsigned maxDegree = _problem.getHGraph().getMaxNodeDegree();  
    int maxGain = maxDegree * 
                  ( _eval.getMaxCostOfOneNet() - _eval.getMinCostOfOneNet() );
    _gainCo.setMaxGain(maxGain);
}

template<class Evaluator, class GainContainer>
void FMMMbase<Evaluator,GainContainer> :: reinitialize()
{ 
   fill(_lockedNetConfigIds.begin(),_lockedNetConfigIds.end(),0);
   fill(_lockedNets        .begin(),_lockedNets        .end(),false);

   _partLeg.reinitialize(); 

   MoveManagerTemplateBase<Evaluator, GainContainer>::reinitialize();
   _gainCo.reinitialize();
   _gainCo.setPartitionBias(0);
// computeGains should be called by partioner

   const vector<double>& newAreas = _partLeg.getActualBalances();
   for(unsigned k = 0; k<_numParts; k++) _partSol->partArea[k][0] = newAreas[k];
   _moveLog.clear();
}

template<class Evaluator, class GainContainer>
void
FMMMbase<Evaluator,GainContainer>::resetTo(PartitioningSolution& newSol) 
{   
    _partSol = &newSol; 
    _curPart = &(newSol.part); 
   fill(_lockedNetConfigIds.begin(),_lockedNetConfigIds.end(),0);
   fill(_lockedNets        .begin(),_lockedNets        .end(),false);

    _partLeg.resetTo(newSol.part);
   //cout << _partLeg ;
   // if made !=0 moves enforcing *capacity* (not fixed) constraints
   // then go balance further
// if (
//   unsigned numMoves=_partLeg.enforce(_problem.getFixedConstr())
//  )  _partLeg.balance(_problem.getFixedConstr())
    //;
 //  cout << " Legalizer did " << numMoves << " moves" << endl;

    _eval.resetTo(newSol.part);
    initializeGainBuckets();  // uses Max/Min net cost from _eval
    _moveLog.resetTo(newSol.part);
 
   const vector<double>& newAreas = _partLeg.getActualBalances();
   for(unsigned k = 0; k < _numParts; k++)
	_partSol->partArea[k][0] = newAreas[k];

    _gainCo.reinitialize(); 
    _gainCo.setPartitionBias(0);
}

template<class Evaluator, class GainContainer>
void 
FMMMbase<Evaluator,GainContainer>::computeGainsOfModule(unsigned moduleIdx)
{

     unsigned from=(*_curPart)[moduleIdx].lowestNumPart();
     PartitionIds canGoTo=(*_fixedConstr)[moduleIdx];

     const HGFNode& node=_problem.getHGraph().getNodeByIdx(moduleIdx);

     // bool tooLarge = (!_allowCorkingNodes  && _solnLegalAtStart &&
     //                         node.getWeight() >= _tol);
     bool tooLarge = (!_allowCorkingNodes && node.getWeight() >= _tol);

     if (_eval.isNetCut2way())
     {  
        unsigned to=1-from;

	//if(tooLarge) cout<<"Skipping node "<<moduleIdx<<endl;

        if(_lockedModules[moduleIdx] = (!canGoTo[to] || tooLarge))
		return; 		/* assignment */
	     
        int gain=0;
        const NetCut2way& eval=reinterpret_cast<NetCut2way&>(_eval);
        const unsigned short * _tallies=eval.getTallies();

        for(itHGFEdgeLocal e = node.edgesBegin(); e!=node.edgesEnd(); e++)
        {
            unsigned edgeIdx=(*e)->getIndex();
            if      (_tallies[2*edgeIdx+  to]==0)      gain--;
            else if (_tallies[2*edgeIdx+from]==1)      gain++;
        } 
        _gainCo.addElement(moduleIdx,to,gain);
        return;
     }

//       _lockedNetConfigIds[edgeIdx]|=(1<<from);  //not clear how to use this

     bool isMovable=false;
     int  baseCost=_eval.getTotalCost();
     for(unsigned to=0; to!=_numParts; to++)
     {
         if (to==from || !canGoTo[to] || tooLarge) continue;
         _eval.moveModuleTo(moduleIdx,from,to);  // to compute "what if" cost
         int gain=baseCost - _eval.getTotalCost();
         _gainCo.addElement(moduleIdx,to,gain);
         isMovable=true;
         _eval.moveModuleTo(moduleIdx,to,from);  // move the module back
     }
     _lockedModules[moduleIdx]= !isMovable;
}

#endif

#endif // (!defined(_MSC_VER) || defined(_INSIDE_FMMOVEMANAGER_H) )
