/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/






#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

//created by Andrew Caldwell and Igor Markov on 06/02/98
#if (!defined(_MSC_VER) || defined(_INSIDE_FMMOVEMANAGER_H) )

#ifndef _FMMM_CXX_
#define _FMMM_CXX_

#ifndef _INSIDE_FMMOVEMANAGER_H
#include "fmMoveManager.h"
#endif //!defined(_INSIDE_FMMOVEMANAGER_H)

#include "PartEvals/netCut2way.h"

template<class Evaluator, class GainContainer>
void FMMoveManager<Evaluator,GainContainer>::computeGains()
{
  if (!randomized)
  {
    unsigned moduleIdx;
    for(moduleIdx=0; moduleIdx!=_numTerminals+_numMovable; moduleIdx++)
       computeGainsOfModule(moduleIdx); 
  }
  else
  { 
    Permutation randPerm(_numTerminals+_numMovable, Mapping::_Random);
    unsigned moduleIdx;
    for(moduleIdx=0; moduleIdx!=_numTerminals+_numMovable; moduleIdx++)
       computeGainsOfModule(randPerm[moduleIdx]);
  }

  _gainCo.resetAfterGainUpdate();
  abkassert(_gainCo.isConsistent(), "GainCo inconsistent in computeGains");
}

#define FMMMclass  FMMoveManager

#define updateGainOfModule( moduleIdx,  from, to,  deltaGain) \
	 _gainCo.updateGain(moduleIdx,to,deltaGain)

// FMMMclass needs to be #defined in the #including class

template<class Evaluator, class GainContainer>
bool FMMMclass<Evaluator,GainContainer>  ::  pickMoveApplyIt()
{
   if (!pickMove()) return false;  // side effect !
   _gainCo.setPartitionBias(_moveToPart);
   // now apply the move

   updateGains();

   _gainCo.resetAfterGainUpdate();

   PartitionIds& movingIds=(*_curPart)[_movingModuleIdx];

   _moveLog.logMove(_movingModuleIdx,movingIds);

// can be done with movingIds.setToPart(_moveToPart); 
// but that won't support replicated modules 
 {
   movingIds.removeFromPart(_moveFromPart);
   movingIds.addToPart     (_moveToPart);
 }

   _eval.moveModuleTo   (_movingModuleIdx,_moveFromPart,_moveToPart);
   _partLeg.moveModuleTo(_movingModuleIdx,_moveFromPart,_moveToPart);

  abkassert(_gainCo.isConsistent(), "GainCo inconsistent in pickMoveApplyIt");

   return true;
} 

template<class Evaluator, class GainContainer>
bool FMMMclass<Evaluator,GainContainer> :: pickMove()      
/* returns false, if no moves are available */
/* side-effect: stores the chosen move as 3 unsigneds */
/* calls isMoveLegal() and can change gain container, thus not const */
{
    int invalidatedNodesInBucket=0;
    SVGainElement* moveToMake = NULL;
    while((moveToMake = _gainCo.getHighestGainElement()) != NULL)
    {
	_movingModuleIdx = _gainCo.getElementId(*moveToMake);
	_moveToPart      = _gainCo.getElementPartition(*moveToMake);
	_moveFromPart    = (*_curPart)[_movingModuleIdx].lowestNumPart();
              
	if(isMoveLegal() || 
	   (_partLeg.getViolation() == 0 && allowIllegalMoves)) return true;
        else
        {
           if (++invalidatedNodesInBucket < unCorking)
           {
               if (_gainCo.invalidateElement(_moveToPart))
                   invalidatedNodesInBucket=0;
           }
           else
           {
               invalidatedNodesInBucket=0;
	       _gainCo.invalidateBucket(_moveToPart);
           }
        }
  
    }
    return false;//nothing legal to move
}

// FMMMclass needs to be #defined in the #including file

template<class Evaluator, class GainContainer>
void FMMMclass<Evaluator,GainContainer>::updateAffectedModules()
// wrt to _movingModuleIdx, _moveFromPart and _moveToPart
{
  unsigned newLockedIds= 1<<_moveToPart;
  const HGFNode& n = _problem.getHGraph().getNodeByIdx(_movingModuleIdx);
  for(itHGFEdgeLocal  e = n.edgesBegin(); e != n.edgesEnd(); e++)
  {
     const HGFEdge& net=(**e);
     unsigned netId=net.getIndex(); 

     if (_lockedNets[netId]) continue;
     if (_lockedNetConfigIds[netId] & (~newLockedIds)) _lockedNets[netId]=true;
     else _lockedNetConfigIds[netId] |=  newLockedIds; 

     if(_eval.netCostNotAffectedByMove(netId,_moveFromPart,_moveToPart)) 
     continue;

     unsigned oldNetCost = _eval.getNetCost(netId);
 
     for(itHGFNodeLocal gn=net.nodesBegin(); gn!=net.nodesEnd(); gn++)
     {
        unsigned gainingModuleIdx= (*gn)->getIndex();
        if (_lockedModules[gainingModuleIdx]) continue; 
        if (gainingModuleIdx==_movingModuleIdx) continue;//will be locked below

        unsigned gnCurPart=(*_curPart)[gainingModuleIdx].lowestNumPart();
        for(unsigned k=0; k!=_numParts; k++)
        {
           if (k==gnCurPart)  continue;
           _eval.updateNetForMovedModule(netId,_moveFromPart,_moveToPart);
           unsigned deltaNetCost = oldNetCost -_eval.computeCostOfOneNet(netId);

           _eval.updateNetForMovedModule(netId,gnCurPart,k);
           unsigned newWhatIfCost= _eval.computeCostOfOneNet(netId);
           _eval.updateNetForMovedModule(netId,_moveToPart,_moveFromPart);
           unsigned deltaWhatIfCost 
                       = _eval.computeCostOfOneNet(netId) - newWhatIfCost;
	   if(deltaWhatIfCost-deltaNetCost != 0)
               updateGainOfModule(gainingModuleIdx, UINT_MAX, k,
					deltaWhatIfCost-deltaNetCost);
           _eval.updateNetForMovedModule(netId,k,gnCurPart);
        }
     }
   }
   // at this point the total cost should be consistent in _eval
   // all nodes at old locations
}

#endif

#endif //(!defined(_MSC_VER) || defined(_INSIDE_FMMOVEMANAGER_H) )
