/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/




// Created: Dec 1, 1999 by Igor Markov <imarkov@cs.ucla.edu>
// Changed: Nov 30, 2000 by Saurabh Adya <sadya@umich.edu>

#include <iostream.h>
#include <fstream.h>
#include "metaPl.h"
#include "ABKCommon/abkcommon.h"
#include "DB/DB.h"
#include "Capo/capoPlacer.h"
#include "RBPlFromDB/RBPlaceFromDB.h"
#include "NetTopology/hgWDimsFromDB.h"
#include "FixOr/greedyOrientOpt.h"
#include "RowIroning/rowIroning.h"

MetaPlacerParams::MetaPlacerParams(int argc, const char *argv[])
:verb(argc,argv)
{
  NoParams         noParams       (argc,argv);
  BoolParam        helpRequest    ("help",argc,argv);
  BoolParam        helpRequest1   ("h",   argc,argv);
  if (noParams.found() || helpRequest.found() || helpRequest1.found())
  {
    cout <<
    " Available MetaPlacer parameters: " << endl <<
    " -verb <1_1_1,1_2_3 etc>     - verbosity\n"
    " -save           - output will be saved into files with base out \n"
    " -saveXorder     - output will be saved into left2right.ord \n"
//  " -flip           - request cell flipping done after placement\n"
    " -noRowIroning   - skip row ironing"<< endl;
    return;
  }

  save             = BoolParam ("save",argc,argv);
  flip             = true; // BoolParam ("flip",argc,argv);
  skipRowIroning   = BoolParam ("noRowIroning",argc,argv);
  saveXorder       = BoolParam ("saveXorder",argc,argv);
}

ostream& operator<< (ostream& out, const MetaPlacerParams& params)
{
  const char *trueFalse[2]={"false","true"};
  out << "MetaPlacerParams:"
      << "\n  Verbosity:     " << params.verb 
      << "\n  Save:          " << trueFalse[params.save] 
      << "\n  Cell Flipping: " << trueFalse[params.flip]
      << "\n  Row Ironing:   " << trueFalse[!params.skipRowIroning]
      << endl;
  return out;
}

int MetaPlacer::_greetingEtc(int argc, const char *argv[])
{
  NoParams         noParams       (argc,argv);
  BoolParam        helpRequest    ("help",argc,argv);
  BoolParam        helpRequest1   ("h",   argc,argv);

  CapoParameters capoParams(argc,argv);
  RowIroningParameters RIparams(argc, argv);

  if (noParams.found() || helpRequest.found() || helpRequest1.found())
  {
   cout << "  \nOther options: \n"
        << "  Plotting options: \n"
	<< "   -minx,-maxx,-miny,-maxy  (for required range. Default means all)\n"
	<< "   -plotNets        base_fileName\n"
	<< "   -plotNodes       base_fileName\n"
	<< "   -plotNodesWNames base_fileName\n"
	<< "   -plot            base_fileName   (plots nodes and nets)\n"
	<< "   -plotWNames      base_fileName   (the above + node names)\n"
	<< "   -plotSites	  base_fileName   (plots the site map)\n"
	<< "   -plotNodesWSites base_fileName   (the above + nodes)\n"
	<< "   -skipLegal (Use to remove any overlaps in existing placement)\n"
	<< endl;
     cout << "Usage: prog -f filename.aux (also see -help)\n"<<endl;
   return 1;
  }
  return 0;


}

void MetaPlacer::_workWithRBPlace (RBPlacement& rbplace, 
int argc, const char *argv[])
{
 CapoParameters capoParams(argc,argv);
 RowIroningParameters   RIparams(argc, argv); 

 StringParam   plotNets("plotNets", argc, argv);
 StringParam   plotNodes("plotNodes", argc, argv);
 StringParam   plotNodesAndNets("plot", argc, argv);
 StringParam   plotNodesWNames("plotNodesWNames", argc, argv);
 StringParam   plotNodesAndNetsWNames("plotWNames", argc, argv);
 StringParam   plotSites("plotSites", argc, argv);
 StringParam   plotNodesWSites("plotNodesWSites", argc, argv);  
 BoolParam     keepOverlaps    ("skipLegal",argc,argv);
 DoubleParam xmin("xmin",argc,argv);
 DoubleParam xmax("xmax",argc,argv);
 DoubleParam ymin("ymin",argc,argv);
 DoubleParam ymax("ymax",argc,argv);

 double minx  = ( xmin.found() ?  xmin:-DBL_MAX);
 double maxx  = ( xmax.found() ?  xmax:DBL_MAX);
 double miny  = ( ymin.found() ?  ymin:-DBL_MAX);
 double maxy  = ( ymax.found() ?  ymax:DBL_MAX);

 cout << capoParams << endl;
 if(!_params.skipRowIroning)
 {
     cout << " Will do row ironing ... " << endl;
     cout << RIparams;
 }

  Timer capoTimer;
  CapoPlacer capo(rbplace,capoParams);
  capoTimer.stop();

  cout << "After Capo  "<<endl;
  cout << "  Capo Runtime: "<<capoTimer.getUserTime()<<endl;

  cout << "  RBPl Est HPWL: "<<rbplace.evalHPWL(false)<<endl;;
  cout << "  RBPl Est Pin-to-Pin HPWL: "<<rbplace.evalHPWL(true)<<endl;;
  cout << "  RBPl Est  WWL: "<<rbplace.evalWeightedWL(false)<<endl;;
  cout << "  RBPl Est Pin-to-Pin  WWL: "<<rbplace.evalWeightedWL(true)<<endl;;

  if(!keepOverlaps.found())
    {
//    rbplace.saveAsPlot("plotWNames",minx,maxx,miny,maxy,"saved");
      rbplace.remOverlaps();
//      rbplace.remOverlapsVert();
    }

  if(_params.flip)
 {
  if (!_db)  
    abkwarn(0,"Cannot [yet] optimize cell orientation without LEF/DEF input\n")
  else
  { 
      cout << " ====== Optimizing cell orientations ... " << endl;
      _db->setPlaceAndOrient(rbplace);
      OrientOptParameters flipParams;

      Timer flipTimer;
      GreedyOrientOpt flipper(*_db,flipParams);
      flipTimer.stop();

      *_db<<flipper;

      cout<<"After Greedy Flipping"<<endl;

      cout << "  Flip Runtime: "<<flipTimer.getUserTime()<<endl;
      cout << "  Total Cells flipped: "<<flipper.getNumFlips()<<endl;
      cout << "  DB   Est Pin-to-Pin HPWL: "<<
        _db->getNetlist().evalHalfPerimCost(_db->spatial.locations,
                                  _db->spatial.orientations)<<endl<<endl;
      rbplace.resetPlacementWOri(
                PlacementWOrient(_db->spatial.locations,
                                 _db->spatial.orientations));     
  }
 }

  if(!_params.skipRowIroning)
  {
     cout << " ====== Launching row ironing ... " << endl;
     cout << RIparams;
     Timer RITimer;
     ironRows(rbplace,RIparams);
     RITimer.stop();

     cout << "After RowIroning  "<<endl;
     cout << "  RowIroning Runtime: "<<RITimer.getUserTime()<<endl;
     cout << "  RBPl Est HPWL: "<<rbplace.evalHPWL(false)<<endl;
     cout << "  RBPl Est Pin-to-Pin HPWL: "<<rbplace.evalHPWL(true)<<endl;
     cout << "  RBPl Est  WWL: "<<rbplace.evalWeightedWL(false)<<endl;
     cout <<"  RBPl Est Pin-to-Pin  WWL: "<<rbplace.evalWeightedWL(true)<<endl;
  }          

  if(_params.save)
  {
    if (_db) 
      {
	cout << "Saving out.def ... " << flush;
	_db->setPlaceAndOrient(rbplace);
	_db->saveDEF("out.def");
      }
    cout << "Saving out.pl ..." << flush;
    rbplace.savePlacement("out.pl");
    cout << "ok" << endl;
  }

  if (_params.saveXorder)
  {
    unsigned idx=0, numCells=rbplace.getNumCells();
    vector<unsigned> toSort(numCells);
    for(;idx!=numCells;idx++) toSort[idx]=idx;
    CompareCellIdsByX byX(rbplace);
    sort(toSort.begin(),toSort.end(),byX);
    const HGraphWDimensions& hgr=rbplace.getHGraph();
    ofstream of("left2right.ord");
    cout << "Saving left2right.ord ... "<< flush;
    for(idx=0;idx!=numCells;idx++)
       of << setw(12) << hgr.getNodeByIdx(toSort[idx]).getName() << endl;
    cout << "ok" << endl;
  }

   if(plotNets.found())
     rbplace.saveAsPlot("plotNets",minx,maxx,miny,maxy,plotNets);

   if(plotNodes.found())
     rbplace.saveAsPlot("plotNodes",minx,maxx,miny,maxy,plotNodes);

   if(plotNodesAndNets.found())
     rbplace.saveAsPlot("plot",minx,maxx,miny,maxy,plotNodesAndNets);

   if(plotNodesWNames.found())
     rbplace.saveAsPlot("plotNodesWNames",minx,maxx,miny,maxy,plotNodesWNames);

   if(plotNodesAndNetsWNames.found())
     rbplace.saveAsPlot("plotWNames",minx,maxx,miny,maxy,plotNodesAndNetsWNames);

   if(plotSites.found())
     rbplace.saveAsPlot("plotSites",minx,maxx,miny,maxy,plotSites);
   
   if(plotNodesWSites.found())
     rbplace.saveAsPlot("plotNodesWSites",minx,maxx,miny,maxy,plotNodesWSites);
 
  return;
}

MetaPlacer::MetaPlacer(int argc, const char *argv[])
:_params(argc,argv),_db(NULL)
{
 if (_greetingEtc(argc,argv)) return;	

 StringParam    auxFileName    ("f",   argc,argv);
 abkfatal(auxFileName.found(),"Usage: prog -f filename.aux \n"
                              " See -help for more params");
 //  figure out what kind of AUX file this is
 //  Three cases are possible: DB, RBPlacement and Hgraph (for linear pl)
 
 ifstream ifs(auxFileName);
 abkfatal3(ifs,"Cannot open ",auxFileName,"\n "); 
 unsigned linesRead=0;
 while (!ifs.eof())
 {  
    if (linesRead>=10)
    {
       cerr << "\n No recognized format spec in first 10 lines of file\n    "
            << auxFileName << endl;
       cerr << " Currently supported formats are LEFDEF*, RowBasedPlacement and HGraph*"
            << endl;
       abkfatal(0," ");
    }
    char buf[1000];
    ifs >> buf;
    ifs >> skiptoeol;
    linesRead++;
    if (buf[0]=='#') continue;
    buf[999]='\0';
    if (strstr(buf,"LEFDEF") || strstr(buf,"lefdef"))
	 { // case 1: DB
           cout << " Got LEF/DEF on input ..." << endl;
	   _db=new DB(auxFileName);
	   cout<<"Design Name: "<<_db->getDesignName()<<endl;
	   RBPlaceParams rbParam(argc,argv);
	   if(rbParam.remCongestion)
	     _db->routedNets.readLEFDEF(_db->getNetlist(), _db->getLayout(), 
				       auxFileName);
	   RBPlaceFromDB rbplace(*_db, rbParam); 
	   _workWithRBPlace(rbplace,argc,argv); 
	   return;
	 }
     else
     if (!strcasecmp(buf,"RowBasedPlacement") || 
         !strcasecmp(buf,"RBPlace") )
	 { // case 2: RBPlace
           cout << " Got RowBasedPlacement on input ..." << endl;
	   RBPlaceParams rbParam(argc,argv);
	   RBPlacement rbplace(auxFileName, rbParam);   
	   _db=NULL;
	   _workWithRBPlace(rbplace,argc,argv);
	   return; 
	 }
     else if (strstr(buf,"raph") || strstr(buf,"RAPH"))
	 { // case 3: HGraph
           cout << " Got HGraph on input ..." << endl;
	   RBPlaceParams rbParam(argc,argv);
	   RBPlacement rbplace(auxFileName, 0.0, 0.0, rbParam);   
	   _db=NULL;
	   _workWithRBPlace(rbplace,argc,argv);
	   return; 
	 }
   }
  cerr << "\n No recognized format spec in first 10 lines of file\n    "
       << auxFileName << endl;
  cerr << " Currently supported formats are LEFDEF* and RowBasedPlacement"
       << endl;
  abkfatal(0," ");
}

MetaPlacer::MetaPlacer(RBPlacement& rbplace, int argc, const char *argv[])
:_params(argc,argv),_db(NULL)
{
 _workWithRBPlace(rbplace,argc,argv); 
}

MetaPlacer::~MetaPlacer()
{
  if (_db) delete _db;
  _db=NULL;
}
