/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/












#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <iostream.h>
#include "stripNetStacks.h"

StripNetStacks::StripNetStacks(
		const SmallPlacementProblem& problem,
		EdgeAndOffset**& firstNets)
	: _numNets(problem.getNumNets()), 
	  _numCells(problem.getNumCells()),
	  _firstNets(firstNets),
	  _totalWidth(0),
	  _bloatWidths(problem.getNumCells())
{

    _numPins  = _numNets*(_numCells+1);

    _netStacks = new XInterval [_numNets] ; //pointers to current net costs

    _cost = 0. ;

    unsigned c; //, n ;
    for(c = 0; c != problem.getNumCells(); c++)
        _totalWidth += problem.getCellWidth(c);

    const SmallPlacementRow& row = problem.getRows()[0] ;
    _leftEdge  = row.xMin ;

    double space = row.xMax - row.xMin ;
    double bloat = (space-_totalWidth)/_numCells ;

    for(c = 0; c < problem.getNumCells(); c++)
        _bloatWidths[c] = problem.getCellWidth(c)+bloat ;

    _nMoveable.reserve(_numNets) ;

    _toLeft.reset() ;

    populate(problem);
}

void StripNetStacks::populate(const SmallPlacementProblem& problem)
{

    double   initCost = 0;
//  unsigned nextEdgeStack = 0;

    XInterval rowIvl(problem.getRows()[0].xMin,problem.getRows()[0].xMax);

    unsigned n, c ;
    for (n = 0 ; n < problem.getNumNets() ; n++) 
    {
        _nMoveable[n] = 0 ;
    }

    const SmallPlacementNetlist& netlist = problem.getNetlist() ;

    unsigned idx = 0 ;
    for (c = 0; c < problem.getNumCells(); c++)
    {
        for (n = 0 ; n < _numNets ; n++)
   	{
	    double pinOffset = netlist.getPinOffset(c,n).x;
            if(pinOffset != DBL_MAX)	//they are connected...
	    {
	        _nMoveable[n]++ ;
	    }
	}
    }

    idx = 0 ;
    for(n = 0; n < problem.getNumNets(); n++)
    {
	const BBox& netBBox = problem.getNetTerminalBBox(n) ;
	XInterval& netIvl = _netStacks[n] ;
	netIvl = XInterval(netBBox.xMin, netBBox.xMax) ;

	if ( ! netIvl.isEmpty() )
	{
            double  rmin=rowIvl.min;
            double& nmax=netIvl.max;
	    if ( nmax < rmin ) nmax = rmin; //same as += rowInv.min
            else if ( netIvl.min > rowIvl.max )
               {netIvl.min = rowIvl.max; }//same as += rowInv.max

	    initCost += netIvl.getLength() ;
	}
    }
    _cost = initCost ;
 
    for (n = 0 ; n < _numNets ; n++)
    {
        const XInterval& netIvl = _netStacks[n];
	if ((! netIvl.isEmpty()) && (_nMoveable[n] != 0))
	{
            if (netIvl.min <= _leftEdge || netIvl.max <= _leftEdge)
                {_toLeft.set(n);} //either end of net <= left edge means trailer.
	}
    }
}

StripNetStacks::~StripNetStacks()
{
    delete[] _netStacks ;
}

void StripNetStacks::print()
{
    cout<<"Current WL for each net is:"<<endl;

    unsigned n ;
    for(n = 0; n < _numNets; n++) 
	cout << setw(2) << n << " : " << setw(3) << _nMoveable[n]
             << " : " << _netStacks[n].getLength() << endl ;

    cout << " bits: " << _toLeft << " " ;
    cout << " cost: " << _cost << endl ;
}

void StripNetStacks::copy(const StripNetStacks& t)
{
    _leftEdge  = t._leftEdge ;

    _cost   = t._cost ;
    _toLeft = t._toLeft ;

    unsigned i ;
    for (i = 0 ; i < _numNets ; i++) {
        _nMoveable[i]     = t._nMoveable[i] ;
        _netStacks[i]     = t._netStacks[i] ;
    }

    //print() ;
}
